@@ -24,10 +24,12 @@
 
 #include "sanlock_internal.h"
 #include "diskio.h"
+#include "ondisk.h"
 #include "direct.h"
 #include "log.h"
 #include "paxos_lease.h"
 #include "delta_lease.h"
+#include "timeouts.h"
 
 /* Based on "Light-Weight Leases for Storage-Centric Coordination"
    by Gregory Chockler and Dahlia Malkhi */
@@ -79,9 +81,9 @@ static int verify_leader(struct sync_disk *disk,
 			 char *space_name,
 			 uint64_t host_id,
 			 struct leader_record *lr,
+			 uint32_t checksum,
 			 const char *caller)
 {
-	uint32_t sum;
 	int result;
 
 	if (lr->magic != DELTA_DISK_MAGIC) {
@@ -116,12 +118,10 @@ static int verify_leader(struct sync_disk *disk,
 		goto fail;
 	}
 
-	sum = leader_checksum(lr);
-
-	if (lr->checksum != sum) {
+	if (lr->checksum != checksum) {
 		log_error("verify_leader %llu wrong checksum %x %x %s",
 			  (unsigned long long)host_id,
-			  lr->checksum, sum, disk->path);
+			  lr->checksum, checksum, disk->path);
 		result = SANLK_LEADER_CHECKSUM;
 		goto fail;
 	}
@@ -132,14 +132,17 @@ static int verify_leader(struct sync_disk *disk,
 	log_leader_error(result, space_name, host_id, disk, lr, caller);
 
 	/*
+	struct leader_record leader_end;
 	struct leader_record leader_rr;
 	int rv;
 
-	memset(&leader_rr, 0, sizeof(leader_rr));
+	memset(&leader_end, 0, sizeof(leader_end));
 
-	rv = read_sectors(disk, host_id - 1, 1, (char *)&leader_rr,
+	rv = read_sectors(disk, host_id - 1, 1, (char *)&leader_end,
 			  sizeof(struct leader_record),
 			  NULL, "delta_verify");
+	
+	leader_record_in(&leader_end, &leader_rr);
 
 	log_leader_error(rv, space_name, host_id, disk, &leader_rr, "delta_verify");
 	*/
@@ -147,32 +150,109 @@ static int verify_leader(struct sync_disk *disk,
 	return result;
 }
 
-int delta_lease_leader_read(struct task *task,
+
+/* read the lockspace name and io_timeout given the disk location */
+
+int delta_read_lockspace(struct task *task,
+			 struct sync_disk *disk,
+			 uint64_t host_id,
+			 struct sanlk_lockspace *ls,
+			 int io_timeout,
+			 int *io_timeout_ret)
+{
+	struct leader_record leader_end;
+	struct leader_record leader;
+	uint32_t checksum;
+	char *space_name;
+	int rv, error;
+
+	/* host_id N is block offset N-1 */
+
+	memset(&leader_end, 0, sizeof(struct leader_record));
+
+	rv = read_sectors(disk, host_id - 1, 1, (char *)&leader_end, sizeof(struct leader_record),
+			  task, io_timeout, "read_lockspace");
+	if (rv < 0)
+		return rv;
+
+	/* N.B. compute checksum before byte swapping */
+	checksum = leader_checksum(&leader_end);
+
+	leader_record_in(&leader_end, &leader);
+
+	if (!ls->name[0])
+		space_name = leader.space_name;
+	else
+		space_name = ls->name;
+
+	error = verify_leader(disk, space_name, host_id, &leader, checksum, "read_lockspace");
+
+	if (error == SANLK_OK) {
+		memcpy(ls->name, leader.space_name, SANLK_NAME_LEN);
+		ls->host_id = host_id;
+		*io_timeout_ret = leader.io_timeout;
+	}
+
+	return error;
+}
+
+int delta_lease_leader_read(struct task *task, int io_timeout,
 			    struct sync_disk *disk,
 			    char *space_name,
 			    uint64_t host_id,
 			    struct leader_record *leader_ret,
 			    const char *caller)
 {
+	struct leader_record leader_end;
 	struct leader_record leader;
+	uint32_t checksum;
 	int rv, error;
 
 	/* host_id N is block offset N-1 */
 
-	memset(&leader, 0, sizeof(struct leader_record));
+	memset(&leader_end, 0, sizeof(struct leader_record));
 	memset(leader_ret, 0, sizeof(struct leader_record));
 
-	rv = read_sectors(disk, host_id - 1, 1, (char *)&leader, sizeof(struct leader_record),
-			  task, "delta_leader");
+	rv = read_sectors(disk, host_id - 1, 1, (char *)&leader_end, sizeof(struct leader_record),
+			  task, io_timeout, "delta_leader");
 	if (rv < 0)
 		return rv;
 
-	error = verify_leader(disk, space_name, host_id, &leader, caller);
+	/* N.B. compute checksum before byte swapping */
+	checksum = leader_checksum(&leader_end);
+
+	leader_record_in(&leader_end, &leader);
+
+	error = verify_leader(disk, space_name, host_id, &leader, checksum, caller);
 
 	memcpy(leader_ret, &leader, sizeof(struct leader_record));
 	return error;
 }
 
+/*
+ * NB. this should not be used to write the leader record, it is meant only
+ * for manually clobbering the disk to corrupt it for testing, or to manually
+ * repair it after it's corrupted.
+ */
+
+int delta_lease_leader_clobber(struct task *task, int io_timeout,
+			       struct sync_disk *disk,
+			       uint64_t host_id,
+			       struct leader_record *leader,
+			       const char *caller)
+{
+	struct leader_record leader_end;
+	int rv;
+
+	leader_record_out(leader, &leader_end);
+
+	rv = write_sector(disk, host_id - 1, (char *)&leader_end, sizeof(struct leader_record),
+			  task, io_timeout, caller);
+	if (rv < 0)
+		return rv;
+	return SANLK_OK;
+}
+
 /*
  * delta_lease_acquire:
  * set the owner of host_id to our_host_name.
@@ -198,22 +278,37 @@ int delta_lease_acquire(struct task *task,
 {
 	struct leader_record leader;
 	struct leader_record leader1;
+	struct leader_record leader_end;
 	uint64_t new_ts;
+	uint32_t checksum;
+	int other_io_timeout, other_host_dead_seconds, other_id_renewal_seconds;
 	int i, error, rv, delay, delta_large_delay;
 
-	log_space(sp, "delta_acquire %llu begin", (unsigned long long)host_id);
+	log_space(sp, "delta_acquire begin %.48s:%llu",
+		  sp->space_name, (unsigned long long)host_id);
 
-	error = delta_lease_leader_read(task, disk, space_name, host_id, &leader,
+	error = delta_lease_leader_read(task, sp->io_timeout, disk, space_name, host_id, &leader,
 					"delta_acquire_begin");
-	if (error < 0)
+	if (error < 0) {
+		log_space(sp, "delta_acquire leader_read1 error %d", error);
 		return error;
+	}
+
+	other_io_timeout = leader.io_timeout;
+
+	if (!other_io_timeout) {
+		log_erros(sp, "delta_acquire use own io_timeout %d", sp->io_timeout);
+		other_io_timeout = sp->io_timeout;
+	} else if (other_io_timeout != sp->io_timeout) {
+		log_erros(sp, "delta_acquire other_io_timeout %u our %u",
+			  leader.io_timeout, sp->io_timeout);
+	}
 
 	if (leader.timestamp == LEASE_FREE)
 		goto write_new;
 
 	if (!strncmp(leader.resource_name, our_host_name, NAME_ID_SIZE)) {
-		log_space(sp, "delta_acquire %llu fast reacquire",
-			  (unsigned long long)host_id);
+		log_space(sp, "delta_acquire fast reacquire");
 		goto write_new;
 	}
 
@@ -235,30 +330,45 @@ int delta_lease_acquire(struct task *task,
 	 * we use here must be the max of the delta delay (D+6d) and
 	 * host_dead_seconds */
 
-	delay = task->host_dead_seconds;
-	delta_large_delay = task->id_renewal_seconds + (6 * task->io_timeout_seconds);
+	/*
+	 * delay = task->host_dead_seconds;
+	 * delta_large_delay = task->id_renewal_seconds + (6 * task->io_timeout_seconds);
+	 * if (delta_large_delay > delay)
+	 * 	delay = delta_large_delay;
+	 */
+
+	other_host_dead_seconds = calc_host_dead_seconds(other_io_timeout);
+	other_id_renewal_seconds = calc_id_renewal_seconds(other_io_timeout);
+
+	delay = other_host_dead_seconds;
+	delta_large_delay = other_id_renewal_seconds + (6 * other_io_timeout);
 	if (delta_large_delay > delay)
 		delay = delta_large_delay;
 
 	while (1) {
 		memcpy(&leader1, &leader, sizeof(struct leader_record));
 
-		log_space(sp, "delta_acquire %llu delta_large_delay %d delay %d",
-			  (unsigned long long)host_id, delta_large_delay, delay);
+		log_space(sp, "delta_acquire delta_large_delay %d delay %d",
+			  delta_large_delay, delay);
 
 		/* TODO: we could reread every several seconds to see if
 		   it has changed, so we can abort more quickly if so */
 
 		for (i = 0; i < delay; i++) {
-			if (sp->external_remove || external_shutdown)
+			if (sp->external_remove || external_shutdown) {
+				log_space(sp, "delta_acquire abort1 remove %d shutdown %d",
+					  sp->external_remove, external_shutdown);
 				return SANLK_ERROR;
+			}
 			sleep(1);
 		}
 
-		error = delta_lease_leader_read(task, disk, space_name, host_id,
+		error = delta_lease_leader_read(task, sp->io_timeout, disk, space_name, host_id,
 						&leader, "delta_acquire_wait");
-		if (error < 0)
+		if (error < 0) {
+			log_space(sp, "delta_acquire leader_read2 error %d", error);
 			return error;
+		}
 
 		if (!memcmp(&leader1, &leader, sizeof(struct leader_record)))
 			break;
@@ -266,7 +376,7 @@ int delta_lease_acquire(struct task *task,
 		if (leader.timestamp == LEASE_FREE)
 			break;
 
-		log_erros(sp, "delta_acquire %llu busy %llu %llu %llu %.48s",
+		log_erros(sp, "delta_acquire host_id %llu busy1 %llu %llu %llu %.48s",
 			  (unsigned long long)host_id,
 			  (unsigned long long)leader.owner_id,
 			  (unsigned long long)leader.owner_generation,
@@ -278,42 +388,57 @@ int delta_lease_acquire(struct task *task,
  write_new:
 	new_ts = monotime();
 	leader.timestamp = new_ts;
+	leader.io_timeout = (sp->io_timeout & 0x00FF);
 	leader.owner_id = host_id;
 	leader.owner_generation++;
 	snprintf(leader.resource_name, NAME_ID_SIZE, "%s", our_host_name);
-	leader.checksum = leader_checksum(&leader);
+	leader.checksum = 0; /* set below */
 
-	log_space(sp, "delta_acquire %llu write %llu %llu %llu %.48s",
-		  (unsigned long long)host_id,
+	log_space(sp, "delta_acquire write %llu %llu %llu %.48s",
 		  (unsigned long long)leader.owner_id,
 		  (unsigned long long)leader.owner_generation,
 		  (unsigned long long)leader.timestamp,
 		  leader.resource_name);
 
-	rv = write_sector(disk, host_id - 1, (char *)&leader, sizeof(struct leader_record),
-			  task, "delta_leader");
-	if (rv < 0)
+	leader_record_out(&leader, &leader_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = leader_checksum(&leader_end);
+	leader.checksum = checksum;
+	leader_end.checksum = cpu_to_le32(checksum);
+
+	rv = write_sector(disk, host_id - 1, (char *)&leader_end, sizeof(struct leader_record),
+			  task, sp->io_timeout, "delta_leader");
+	if (rv < 0) {
+		log_space(sp, "delta_acquire write error %d", rv);
 		return rv;
+	}
 
 	memcpy(&leader1, &leader, sizeof(struct leader_record));
 
-	delay = 2 * task->io_timeout_seconds;
-	log_space(sp, "delta_acquire %llu delta_short_delay %d",
-		  (unsigned long long)host_id, delay);
+	delay = 2 * other_io_timeout;
+	log_space(sp, "delta_acquire delta_short_delay %d", delay);
 
 	for (i = 0; i < delay; i++) {
-		if (sp->external_remove || external_shutdown)
+		if (sp->external_remove || external_shutdown) {
+			log_space(sp, "delta_acquire abort2 remove %d shutdown %d",
+				  sp->external_remove, external_shutdown);
 			return SANLK_ERROR;
+		}
 		sleep(1);
 	}
 
-	error = delta_lease_leader_read(task, disk, space_name, host_id, &leader,
+	error = delta_lease_leader_read(task, sp->io_timeout, disk, space_name, host_id, &leader,
 					"delta_acquire_check");
-	if (error < 0)
+	if (error < 0) {
+		log_space(sp, "delta_acquire leader_read3 error %d", error);
 		return error;
+	}
 
 	if (memcmp(&leader1, &leader, sizeof(struct leader_record))) {
-		log_erros(sp, "delta_acquire %llu busy %llu %llu %llu %.48s",
+		log_erros(sp, "delta_acquire host_id %llu busy2 %llu %llu %llu %.48s",
 			  (unsigned long long)host_id,
 			  (unsigned long long)leader.owner_id,
 			  (unsigned long long)leader.owner_generation,
@@ -322,6 +447,11 @@ int delta_lease_acquire(struct task *task,
 		return SANLK_HOSTID_BUSY;
 	}
 
+	log_space(sp, "delta_acquire done %llu %llu %llu",
+		  (unsigned long long)leader.owner_id,
+		  (unsigned long long)leader.owner_generation,
+		  (unsigned long long)leader.timestamp);
+
 	memcpy(leader_ret, &leader, sizeof(struct leader_record));
 	return SANLK_OK;
 }
@@ -331,20 +461,27 @@ int delta_lease_renew(struct task *task,
 		      struct sync_disk *disk,
 		      char *space_name,
 		      char *bitmap,
+		      struct delta_extra *extra,
 		      int prev_result,
 		      int *read_result,
+		      int log_renewal_level,
 		      struct leader_record *leader_last,
 		      struct leader_record *leader_ret)
 {
 	struct leader_record leader;
+	struct leader_record leader_end;
 	char **p_iobuf;
 	char **p_wbuf;
 	char *wbuf;
-	uint64_t host_id, id_offset, new_ts;
-	int rv, iobuf_len, sector_size, io_timeout_save;
+	uint32_t checksum;
+	uint32_t reap_timeout_msec;
+	uint64_t host_id, id_offset, new_ts, now;
+	int rv, iobuf_len, sector_size;
 
-	if (!leader_last)
+	if (!leader_last) {
+		log_erros(sp, "delta_renew no leader_last");
 		return -EINVAL;
+	}
 
 	*read_result = SANLK_ERROR;
 
@@ -356,9 +493,11 @@ int delta_lease_renew(struct task *task,
 
 	/* offset of our leader_record */
 	id_offset = (host_id - 1) * sector_size;
-	if (id_offset > iobuf_len)
+	if (id_offset > iobuf_len) {
+		log_erros(sp, "delta_renew bad offset %llu iobuf_len %d",
+			  (unsigned long long)id_offset, iobuf_len);
 		return -EINVAL;
-
+	}
 
 	/* if the previous renew timed out in this initial read, and that read
 	   is now complete, we can use that result here instead of discarding
@@ -378,8 +517,18 @@ int delta_lease_renew(struct task *task,
 			goto skip_reap;
 		}
 
+		log_space(sp, "delta_renew begin reap");
+
+		if (!sp->renewal_read_extend_sec) {
+			/* only wait .5 sec when trying to reap a prev io to clear it */
+			reap_timeout_msec = 500;
+		} else {
+			/* effectively continue/extend the read phase from the previous renewal */
+			reap_timeout_msec = sp->renewal_read_extend_sec * 1000;
+		}
+
 		rv = read_iobuf_reap(disk->fd, disk->offset,
-				     task->iobuf, iobuf_len, task);
+				     task->iobuf, iobuf_len, task, reap_timeout_msec);
 
 		log_space(sp, "delta_renew reap %d", rv);
 
@@ -411,6 +560,12 @@ int delta_lease_renew(struct task *task,
 		task->iobuf = NULL;
 	}
 
+	/*
+	 * NB. this task->iobuf is also copied by the lockspace thread
+	 * into renewal_read_buf, which is then copied in the main loop
+	 * by check_our_lease and passed to check_other_leases.
+	 */
+
 	if (!task->iobuf) {
 		/* this will happen the first time renew is called, and after
 		   a timed out renewal read fails to be reaped (see
@@ -425,24 +580,38 @@ int delta_lease_renew(struct task *task,
 		}
 	}
 
-	rv = read_iobuf(disk->fd, disk->offset, task->iobuf, iobuf_len, task);
+	if (log_renewal_level != -1)
+		log_level(sp->space_id, 0, NULL, log_renewal_level, "delta_renew begin read");
+
+	rv = read_iobuf(disk->fd, disk->offset, task->iobuf, iobuf_len, task, sp->io_timeout);
 	if (rv) {
 		/* the next time delta_lease_renew() is called, prev_result
 		   will be this rv.  If this rv is SANLK_AIO_TIMEOUT, we'll
 		   try to reap the event */
 
-		log_erros(sp, "delta_renew read rv %d offset %llu %s",
-			  rv, (unsigned long long)disk->offset, disk->path);
+		if (rv == SANLK_AIO_TIMEOUT)
+			log_erros(sp, "delta_renew read timeout %u sec offset %llu %s",
+				  sp->io_timeout, (unsigned long long)disk->offset, disk->path);
+		else
+			log_erros(sp, "delta_renew read rv %d offset %llu %s",
+				  rv, (unsigned long long)disk->offset, disk->path);
 		return rv;
 	}
 
  read_done:
 	*read_result = SANLK_OK;
-	memcpy(&leader, task->iobuf+id_offset, sizeof(struct leader_record));
+	memcpy(&leader_end, task->iobuf+id_offset, sizeof(struct leader_record));
 
-	rv = verify_leader(disk, space_name, host_id, &leader, "delta_renew");
-	if (rv < 0)
+	/* N.B. compute checksum before byte swapping */
+	checksum = leader_checksum(&leader_end);
+
+	leader_record_in(&leader_end, &leader);
+
+	rv = verify_leader(disk, space_name, host_id, &leader, checksum, "delta_renew");
+	if (rv < 0) {
+		log_erros(sp, "delta_renew verify_leader error %d", rv);
 		return rv;
+	}
 
 	/* We can't always memcmp(&leader, leader_last) because previous writes
 	   may have timed out and we don't know if they were actually written
@@ -466,14 +635,29 @@ int delta_lease_renew(struct task *task,
 		return SANLK_RENEW_DIFF;
 	}
 
-	new_ts = monotime();
-
-	if (leader.timestamp >= new_ts) {
-		log_erros(sp, "delta_renew timestamp too small");
+	if (leader.io_timeout != sp->io_timeout) {
+		log_erros(sp, "delta_renew io_timeout changed disk %d sp %d",
+			  leader.io_timeout, sp->io_timeout);
+		leader.io_timeout = (sp->io_timeout & 0x00FF);
 	}
 
+	new_ts = monotime();
+
+	if (log_renewal_level != -1)
+		log_level(sp->space_id, 0, NULL, log_renewal_level, "delta_renew begin write for new ts %llu", (unsigned long long)new_ts);
+
+	if (leader.timestamp >= new_ts)
+		log_erros(sp, "delta_renew timestamp too small");
+
 	leader.timestamp = new_ts;
-	leader.checksum = leader_checksum(&leader);
+	leader.checksum = 0; /* set below */
+
+	/* TODO: rename the leader fields */
+	if (extra) {
+		leader.write_id = extra->field1;
+		leader.write_generation = extra->field2;
+		leader.write_timestamp = extra->field3;
+	}
 
 	p_wbuf = &wbuf;
 	rv = posix_memalign((void *)p_wbuf, getpagesize(), sector_size);
@@ -482,7 +666,17 @@ int delta_lease_renew(struct task *task,
 		return -ENOMEM;
 	}
 	memset(wbuf, 0, sector_size);
-	memcpy(wbuf, &leader, sizeof(struct leader_record));
+
+	leader_record_out(&leader, &leader_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = leader_checksum(&leader_end);
+	leader.checksum = checksum;
+	leader_end.checksum = cpu_to_le32(checksum);
+
+	memcpy(wbuf, &leader_end, sizeof(struct leader_record));
 	memcpy(wbuf+LEADER_RECORD_MAX, bitmap, HOSTID_BITMAP_SIZE);
 
 	/* extend io timeout for this one write; we need to give this write
@@ -490,18 +684,23 @@ int delta_lease_renew(struct task *task,
 	   out.  there's nothing we would do but retry it, and timing out and
 	   retrying unnecessarily would probably be counter productive. */
 
-	io_timeout_save = task->io_timeout_seconds;
-	task->io_timeout_seconds = task->host_dead_seconds;
-
-	rv = write_iobuf(disk->fd, disk->offset+id_offset, wbuf, sector_size, task);
+	rv = write_iobuf(disk->fd, disk->offset+id_offset, wbuf, sector_size, task,
+			 calc_host_dead_seconds(sp->io_timeout));
 
 	if (rv != SANLK_AIO_TIMEOUT)
 		free(wbuf);
 
-	task->io_timeout_seconds = io_timeout_save;
+	now = monotime();
 
-	if (rv < 0)
+	if (rv < 0) {
+		log_erros(sp, "delta_renew write time %llu error %d",
+			  (unsigned long long)(now - new_ts), rv);
 		return rv;
+	}
+
+	if (now - new_ts >= sp->io_timeout)
+		log_erros(sp, "delta_renew long write time %llu sec",
+			  (unsigned long long)(now - new_ts));
 
 	/* the paper shows doing a delay and another read here, but it seems
 	   unnecessary since we do the same at the beginning of the next renewal */
@@ -518,7 +717,9 @@ int delta_lease_release(struct task *task,
 			struct leader_record *leader_ret)
 {
 	struct leader_record leader;
+	struct leader_record leader_end;
 	uint64_t host_id;
+	uint32_t checksum;
 	int rv;
 
 	if (!leader_last)
@@ -526,16 +727,33 @@ int delta_lease_release(struct task *task,
 
 	host_id = leader_last->owner_id;
 
-	log_space(sp, "delta_release %llu begin", (unsigned long long)host_id);
+	log_space(sp, "delta_release begin %.48s:%llu",
+		  sp->space_name, (unsigned long long)host_id);
 
 	memcpy(&leader, leader_last, sizeof(struct leader_record));
 	leader.timestamp = LEASE_FREE;
-	leader.checksum = leader_checksum(&leader);
+	leader.checksum = 0; /* set below */
 
-	rv = write_sector(disk, host_id - 1, (char *)&leader, sizeof(struct leader_record),
-			  task, "delta_leader");
-	if (rv < 0)
+	leader_record_out(&leader, &leader_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = leader_checksum(&leader_end);
+	leader.checksum = checksum;
+	leader_end.checksum = cpu_to_le32(checksum);
+
+	rv = write_sector(disk, host_id - 1, (char *)&leader_end, sizeof(struct leader_record),
+			  task, sp->io_timeout, "delta_leader");
+	if (rv < 0) {
+		log_space(sp, "delta_release write error %d", rv);
 		return rv;
+	}
+
+	log_space(sp, "delta_release done %llu %llu %llu",
+		  (unsigned long long)leader.owner_id,
+		  (unsigned long long)leader.owner_generation,
+		  (unsigned long long)leader.timestamp);
 
 	memcpy(leader_ret, &leader, sizeof(struct leader_record));
 	return SANLK_OK;
@@ -545,19 +763,29 @@ int delta_lease_release(struct task *task,
    block device disk->path */
 
 int delta_lease_init(struct task *task,
+		     int io_timeout,
 		     struct sync_disk *disk,
 		     char *space_name,
 		     int max_hosts)
 {
-	struct leader_record *leader;
+	struct leader_record leader_first;
+	struct leader_record leader_end;
+	struct leader_record leader;
 	char *iobuf, **p_iobuf;
 	int iobuf_len;
 	int align_size;
 	int i, rv;
+	uint32_t checksum;
 
 	if (!max_hosts)
 		max_hosts = DEFAULT_MAX_HOSTS;
 
+	if (max_hosts > DEFAULT_MAX_HOSTS)
+		return -E2BIG;
+
+	if (!io_timeout)
+		io_timeout = DEFAULT_IO_TIMEOUT;
+
 	align_size = direct_align(disk);
 	if (align_size < 0)
 		return align_size;
@@ -578,24 +806,60 @@ int delta_lease_init(struct task *task,
 	/* host_id N is block offset N-1 */
 
 	for (i = 0; i < max_hosts; i++) {
-		leader = (struct leader_record *)(iobuf + (i * disk->sector_size));
-		leader->magic = DELTA_DISK_MAGIC;
-		leader->version = DELTA_DISK_VERSION_MAJOR | DELTA_DISK_VERSION_MINOR;
-		leader->sector_size = disk->sector_size;
-		leader->max_hosts = 1;
-		leader->timestamp = LEASE_FREE;
-		strncpy(leader->space_name, space_name, NAME_ID_SIZE);
-		leader->checksum = leader_checksum(leader);
+		memset(&leader, 0, sizeof(struct leader_record));
+		leader.magic = DELTA_DISK_MAGIC;
+		leader.version = DELTA_DISK_VERSION_MAJOR | DELTA_DISK_VERSION_MINOR;
+		leader.sector_size = disk->sector_size;
+		leader.max_hosts = 1;
+		leader.timestamp = LEASE_FREE;
+		leader.io_timeout = io_timeout;
+		strncpy(leader.space_name, space_name, NAME_ID_SIZE);
+		leader.checksum = 0; /* set below */
+
+		/* make the first record invalid so we can do a single atomic
+		   write below to commit the whole thing */
+		if (!i) {
+			leader.magic = 0;
+			memcpy(&leader_first, &leader, sizeof(struct leader_record));
+		}
+
+		leader_record_out(&leader, &leader_end);
+
+		/*
+		 * N.B. must compute checksum after the data has been byte swapped.
+		 */
+		checksum = leader_checksum(&leader_end);
+		leader.checksum = checksum;
+		leader_end.checksum = cpu_to_le32(checksum);
+
+		memcpy(iobuf + (i * disk->sector_size), &leader_end, sizeof(struct leader_record));
 	}
 
-	rv = write_iobuf(disk->fd, disk->offset, iobuf, iobuf_len, task);
+	rv = write_iobuf(disk->fd, disk->offset, iobuf, iobuf_len, task, io_timeout);
+	if (rv < 0)
+		goto out;
 
+	/* commit the whole lockspace by making the first record valid */
+
+	leader_first.magic = DELTA_DISK_MAGIC;
+	leader_first.checksum = 0; /* set below */
+
+	leader_record_out(&leader_first, &leader_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = leader_checksum(&leader_end);
+	leader_first.checksum = checksum;
+	leader_end.checksum = cpu_to_le32(checksum);
+
+	memcpy(iobuf, &leader_end, sizeof(struct leader_record));
+
+	rv = write_iobuf(disk->fd, disk->offset, iobuf, disk->sector_size, task, io_timeout);
+ out:
 	if (rv != SANLK_AIO_TIMEOUT)
 		free(iobuf);
 
-	if (rv < 0)
-		return rv;
-
-	return 0;
+	return rv;
 }
 
