@@ -10,6 +10,7 @@
 #define __DELTA_LEASE_H__
 
 int delta_lease_leader_read(struct task *task,
+			    int io_timeout,
 			    struct sync_disk *disk,
 			    char *space_name,
 			    uint64_t host_id,
@@ -29,8 +30,10 @@ int delta_lease_renew(struct task *task,
                       struct sync_disk *disk,
                       char *space_name,
                       char *bitmap,
+		      struct delta_extra *extra,
                       int prev_result,
 		      int *read_result,
+		      int log_renewal_level,
                       struct leader_record *leader_last,
                       struct leader_record *leader_ret);
 
@@ -42,8 +45,22 @@ int delta_lease_release(struct task *task,
                         struct leader_record *leader_ret);
 
 int delta_lease_init(struct task *task,
+		     int io_timeout,
 		     struct sync_disk *disk,
 		     char *space_name,
 		     int max_hosts);
 
+int delta_read_lockspace(struct task *task,
+			struct sync_disk *disk,
+			uint64_t host_id,
+			struct sanlk_lockspace *ls,
+			int io_timeout,
+			int *io_timeout_ret);
+
+int delta_lease_leader_clobber(struct task *task, int io_timeout,
+                               struct sync_disk *disk,
+                               uint64_t host_id,
+                               struct leader_record *leader,
+                               const char *caller);
+
 #endif
