@@ -23,28 +23,18 @@
 
 #include "sanlock_internal.h"
 #include "diskio.h"
+#include "ondisk.h"
 #include "direct.h"
 #include "log.h"
 #include "lockspace.h"
 #include "delta_lease.h"
 #include "paxos_lease.h"
 #include "resource.h"
+#include "timeouts.h"
 
 uint32_t crc32c(uint32_t crc, uint8_t *data, size_t length);
 int get_rand(int a, int b);
 
-#define DBLOCK_CHECKSUM_LEN	 48  /* ends before checksum field */
-
-struct paxos_dblock {
-	uint64_t mbal;
-	uint64_t bal;
-	uint64_t inp;	/* host_id */
-	uint64_t inp2;	/* host_id generation */
-	uint64_t inp3;	/* host_id's timestamp */
-	uint64_t lver;
-	uint32_t checksum;
-};
-
 static uint32_t roundup_power_of_two(uint32_t val)
 {
 	val--;
@@ -57,80 +47,264 @@ static uint32_t roundup_power_of_two(uint32_t val)
 	return val;
 }
 
+uint32_t leader_checksum(struct leader_record *lr)
+{
+	return crc32c((uint32_t)~1, (uint8_t *)lr, LEADER_CHECKSUM_LEN);
+}
+
+static uint32_t dblock_checksum(struct paxos_dblock *pd)
+{
+	return crc32c((uint32_t)~1, (uint8_t *)pd, DBLOCK_CHECKSUM_LEN);
+}
+
 int paxos_lease_request_read(struct task *task, struct token *token,
 			     struct request_record *rr)
 {
+	struct request_record rr_end;
 	int rv;
 
 	/* 1 = request record is second sector */
 
-	rv = read_sectors(&token->disks[0], 1, 1, (char *)rr,
+	rv = read_sectors(&token->disks[0], 1, 1, (char *)&rr_end,
 			  sizeof(struct request_record),
-			  task, "request");
+			  task, token->io_timeout, "request");
 	if (rv < 0)
 		return rv;
+
+	request_record_in(&rr_end, rr);
+
 	return SANLK_OK;
 }
 
 int paxos_lease_request_write(struct task *task, struct token *token,
 			      struct request_record *rr)
 {
+	struct request_record rr_end;
 	int rv;
 
-	rv = write_sector(&token->disks[0], 1, (char *)rr,
+	request_record_out(rr, &rr_end);
+
+	rv = write_sector(&token->disks[0], 1, (char *)&rr_end,
 			  sizeof(struct request_record),
-			  task, "request");
+			  task, token->io_timeout, "request");
 	if (rv < 0)
 		return rv;
+
 	return SANLK_OK;
 }
 
 static int write_dblock(struct task *task,
+		        struct token *token,
+			struct sync_disk *disk,
+			uint64_t host_id,
+			struct paxos_dblock *pd);
+
+int paxos_erase_dblock(struct task *task,
+		       struct token *token,
+		       uint64_t host_id)
+{
+	struct paxos_dblock dblock_end;
+	int num_disks = token->r.num_disks;
+	int num_writes = 0;
+	int d, rv, error = -1;
+
+	memset(&dblock_end, 0, sizeof(struct paxos_dblock));
+
+	for (d = 0; d < num_disks; d++) {
+		rv = write_dblock(task, token, &token->disks[d], host_id, &dblock_end);
+		if (rv < 0) {
+			error = rv;
+			continue;
+		}
+		num_writes++;
+	}
+
+	if (!majority_disks(num_disks, num_writes))
+		return error;
+	return SANLK_OK;
+}
+
+/*
+ * Write a combined dblock and mblock.  This is an odd case that doesn't fit
+ * well with the way the code has been written.  It's used when we want to
+ * convert sh to ex, which requires acquiring the lease owner, but we don't
+ * want to clobber our SHARED mblock by writing a plain dblock in the process
+ * in case there's a problem with the acquiring, we don't want to loose our
+ * shared mode lease.
+ *
+ * NB. this assumes the only mblock flag we want is MBLOCK_SHARED and that
+ * the generation we want is token->host_generation.  This is currently
+ * the case, but could change in the future.
+ */
+
+static int write_dblock_mblock_sh(struct task *task,
+			          struct token *token,
+			          struct sync_disk *disk,
+			          uint64_t host_id,
+			          struct paxos_dblock *pd)
+{
+	struct paxos_dblock pd_end;
+	struct mode_block mb;
+	struct mode_block mb_end;
+	char *iobuf, **p_iobuf;
+	uint64_t offset;
+	uint32_t checksum;
+	int iobuf_len, rv;
+
+	memset(&mb, 0, sizeof(mb));
+	mb.flags = MBLOCK_SHARED;
+	mb.generation = token->host_generation;
+
+	iobuf_len = disk->sector_size;
+	if (!iobuf_len)
+		return -EINVAL;
+
+	p_iobuf = &iobuf;
+
+	rv = posix_memalign((void *)p_iobuf, getpagesize(), iobuf_len);
+	if (rv)
+		return -ENOMEM;
+
+	offset = disk->offset + ((2 + host_id - 1) * disk->sector_size);
+
+	paxos_dblock_out(pd, &pd_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = dblock_checksum(&pd_end);
+	pd->checksum = checksum;
+	pd_end.checksum = cpu_to_le32(checksum);
+
+	mode_block_out(&mb, &mb_end);
+
+	memcpy(iobuf, (char *)&pd_end, sizeof(struct paxos_dblock));
+	memcpy(iobuf + MBLOCK_OFFSET, (char *)&mb_end, sizeof(struct mode_block));
+
+	rv = write_iobuf(disk->fd, offset, iobuf, iobuf_len, task, token->io_timeout);
+
+	if (rv < 0) {
+		log_errot(token, "write_dblock_mblock_sh host_id %llu gen %llu rv %d",
+			  (unsigned long long)host_id,
+			  (unsigned long long)token->host_generation,
+			  rv);
+	}
+
+	if (rv != SANLK_AIO_TIMEOUT)
+		free(iobuf);
+	return rv;
+}
+
+static int write_dblock(struct task *task,
+		        struct token *token,
 			struct sync_disk *disk,
 			uint64_t host_id,
 			struct paxos_dblock *pd)
 {
+	struct paxos_dblock pd_end;
+	uint32_t checksum;
 	int rv;
 
+	if (token->flags & T_WRITE_DBLOCK_MBLOCK_SH) {
+		/* special case to preserve our SH mode block within the dblock */
+		return write_dblock_mblock_sh(task, token, disk, host_id, pd);
+	}
+
 	/* 1 leader block + 1 request block;
 	   host_id N is block offset N-1 */
 
-	rv = write_sector(disk, 2 + host_id - 1, (char *)pd, sizeof(struct paxos_dblock),
-			  task, "dblock");
+	paxos_dblock_out(pd, &pd_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = dblock_checksum(&pd_end);
+	pd->checksum = checksum;
+	pd_end.checksum = cpu_to_le32(checksum);
+
+	rv = write_sector(disk, 2 + host_id - 1, (char *)&pd_end, sizeof(struct paxos_dblock),
+			  task, token->io_timeout, "dblock");
 	return rv;
 }
 
 static int write_leader(struct task *task,
+		        struct token *token,
 			struct sync_disk *disk,
 			struct leader_record *lr)
 {
+	struct leader_record lr_end;
+	uint32_t checksum;
 	int rv;
 
-	rv = write_sector(disk, 0, (char *)lr, sizeof(struct leader_record),
-			  task, "leader");
+	leader_record_out(lr, &lr_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = leader_checksum(&lr_end);
+	lr->checksum = checksum;
+	lr_end.checksum = cpu_to_le32(checksum);
+
+	rv = write_sector(disk, 0, (char *)&lr_end, sizeof(struct leader_record),
+			  task, token->io_timeout, "leader");
+	return rv;
+}
+
+/*
+ * NB. this should not be used to write the leader record, it is meant only
+ * for manually clobbering the disk to corrupt it for testing, or to manually
+ * repair it after it's corrupted.
+ */
+
+int paxos_lease_leader_clobber(struct task *task,
+			       struct token *token,
+			       struct leader_record *leader,
+			       const char *caller)
+{
+	struct leader_record lr_end;
+	uint32_t checksum;
+	int rv;
+
+	leader_record_out(leader, &lr_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = leader_checksum(&lr_end);
+	leader->checksum = checksum;
+	lr_end.checksum = cpu_to_le32(checksum);
+
+	rv = write_sector(&token->disks[0], 0, (char *)&lr_end, sizeof(struct leader_record),
+			  task, token->io_timeout, caller);
 	return rv;
 }
 
-#if 0
 static int read_dblock(struct task *task,
+		       struct token *token,
 		       struct sync_disk *disk,
 		       uint64_t host_id,
 		       struct paxos_dblock *pd)
 {
+	struct paxos_dblock pd_end;
 	int rv;
 
 	/* 1 leader block + 1 request block; host_id N is block offset N-1 */
 
-	rv = read_sectors(disk, 2 + host_id - 1, 1, (char *)pd, sizeof(struct paxos_dblock),
-			  task, "dblock");
+	rv = read_sectors(disk, 2 + host_id - 1, 1, (char *)&pd_end, sizeof(struct paxos_dblock),
+			  task, token->io_timeout, "dblock");
+
+	paxos_dblock_in(&pd_end, pd);
+
 	return rv;
 }
 
+#if 0
 static int read_dblocks(struct task *task,
 			struct sync_disk *disk,
 			struct paxos_dblock *pds,
 			int pds_count)
 {
+	struct paxos_dblock pd_end;
 	char *data;
 	int data_len, rv, i;
 
@@ -154,8 +328,12 @@ static int read_dblocks(struct task *task,
 	   paxos_dblock */
 
 	for (i = 0; i < pds_count; i++) {
-		memcpy(&pds[i], data + (i * disk->sector_size),
+		memcpy(&pd_end, data + (i * disk->sector_size),
 		       sizeof(struct paxos_dblock));
+
+		paxos_dblock_in(&pd_end, &pd);
+
+		memcpy(&pds[i], &pd, sizeof(struct paxos_dblock));
 	}
 
 	rv = 0;
@@ -167,36 +345,35 @@ static int read_dblocks(struct task *task,
 #endif
 
 static int read_leader(struct task *task,
+		       struct token *token,
 		       struct sync_disk *disk,
-		       struct leader_record *lr)
+		       struct leader_record *lr,
+		       uint32_t *checksum)
 {
+	struct leader_record lr_end;
 	int rv;
 
 	/* 0 = leader record is first sector */
 
-	rv = read_sectors(disk, 0, 1, (char *)lr, sizeof(struct leader_record),
-			  task, "leader");
+	rv = read_sectors(disk, 0, 1, (char *)&lr_end, sizeof(struct leader_record),
+			  task, token->io_timeout, "leader");
+
+	/* N.B. checksum is computed while the data is in ondisk format. */
+	*checksum = leader_checksum(&lr_end);
+
+	leader_record_in(&lr_end, lr);
 
 	return rv;
 }
 
-static uint32_t dblock_checksum(struct paxos_dblock *pd)
+static int verify_dblock(struct token *token, struct paxos_dblock *pd, uint32_t checksum)
 {
-	return crc32c((uint32_t)~1, (uint8_t *)pd, DBLOCK_CHECKSUM_LEN);
-}
-
-static int verify_dblock(struct token *token, struct paxos_dblock *pd)
-{
-	uint32_t sum;
-
 	if (!pd->checksum && !pd->mbal && !pd->bal && !pd->inp && !pd->lver)
 		return SANLK_OK;
 
-	sum = dblock_checksum(pd);
-
-	if (pd->checksum != sum) {
+	if (pd->checksum != checksum) {
 		log_errot(token, "verify_dblock wrong checksum %x %x",
-			  pd->checksum, sum);
+			  pd->checksum, checksum);
 		return SANLK_DBLOCK_CHECKSUM;
 	}
 
@@ -255,17 +432,21 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 		      struct paxos_dblock *dblock_out)
 {
 	struct paxos_dblock dblock;
+	struct paxos_dblock bk_in;
 	struct paxos_dblock bk_max;
+	struct paxos_dblock *bk_end;
 	struct paxos_dblock *bk;
 	struct sync_disk *disk;
 	char *iobuf[SANLK_MAX_DISKS];
 	char **p_iobuf[SANLK_MAX_DISKS];
+	uint32_t checksum;
 	int num_disks = token->r.num_disks;
 	int num_writes, num_reads;
 	int sector_size = token->disks[0].sector_size;
 	int sector_count;
 	int iobuf_len;
-	int d, q, rv;
+	int phase2 = 0;
+	int d, q, rv = 0;
 	int q_max = -1;
 	int error;
 
@@ -303,14 +484,14 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 	memset(&dblock, 0, sizeof(struct paxos_dblock));
 	dblock.mbal = our_mbal;
 	dblock.lver = next_lver;
-	dblock.checksum = dblock_checksum(&dblock);
+	dblock.checksum = 0; /* set after paxos_dblock_out */
 
 	memset(&bk_max, 0, sizeof(struct paxos_dblock));
 
 	num_writes = 0;
 
 	for (d = 0; d < num_disks; d++) {
-		rv = write_dblock(task, &token->disks[d], token->host_id, &dblock);
+		rv = write_dblock(task, token, &token->disks[d], token->host_id, &dblock);
 		if (rv < 0)
 			continue;
 		num_writes++;
@@ -332,22 +513,26 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 			continue;
 		memset(iobuf[d], 0, iobuf_len);
 
-		rv = read_iobuf(disk->fd, disk->offset, iobuf[d], iobuf_len, task);
+		rv = read_iobuf(disk->fd, disk->offset, iobuf[d], iobuf_len, task, token->io_timeout);
 		if (rv == SANLK_AIO_TIMEOUT)
 			iobuf[d] = NULL;
 		if (rv < 0)
 			continue;
 		num_reads++;
 
-
 		for (q = 0; q < num_hosts; q++) {
-			bk = (struct paxos_dblock *)(iobuf[d] + ((2 + q)*sector_size));
+			bk_end = (struct paxos_dblock *)(iobuf[d] + ((2 + q)*sector_size));
 
-			rv = verify_dblock(token, bk);
+			checksum = dblock_checksum(bk_end);
+
+			paxos_dblock_in(bk_end, &bk_in);
+			bk = &bk_in;
+
+			rv = verify_dblock(token, bk, checksum);
 			if (rv < 0)
 				continue;
 
-			check_mode_block(token, q, (char *)bk);
+			check_mode_block(token, next_lver, q, (char *)bk_end);
 
 			if (bk->lver < dblock.lver)
 				continue;
@@ -426,7 +611,7 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 		dblock.inp3 = monotime();
 	}
 	dblock.bal = dblock.mbal;
-	dblock.checksum = dblock_checksum(&dblock);
+	dblock.checksum = 0; /* set after paxos_dblock_out */
 
 	if (bk_max.inp) {
 		/* not a problem, but interesting to see, so use log_error */
@@ -447,6 +632,8 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 	 * Same description as phase 1, same sequence of writes/reads.
 	 */
 
+	phase2 = 1;
+
 	log_token(token, "ballot %llu phase2 bal %llu inp %llu %llu %llu q_max %d",
 		  (unsigned long long)dblock.lver,
 		  (unsigned long long)dblock.bal,
@@ -458,7 +645,7 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 	num_writes = 0;
 
 	for (d = 0; d < num_disks; d++) {
-		rv = write_dblock(task, &token->disks[d], token->host_id, &dblock);
+		rv = write_dblock(task, token, &token->disks[d], token->host_id, &dblock);
 		if (rv < 0)
 			continue;
 		num_writes++;
@@ -480,7 +667,7 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 			continue;
 		memset(iobuf[d], 0, iobuf_len);
 
-		rv = read_iobuf(disk->fd, disk->offset, iobuf[d], iobuf_len, task);
+		rv = read_iobuf(disk->fd, disk->offset, iobuf[d], iobuf_len, task, token->io_timeout);
 		if (rv == SANLK_AIO_TIMEOUT)
 			iobuf[d] = NULL;
 		if (rv < 0)
@@ -488,9 +675,14 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 		num_reads++;
 
 		for (q = 0; q < num_hosts; q++) {
-			bk = (struct paxos_dblock *)(iobuf[d] + ((2 + q)*sector_size));
+			bk_end = (struct paxos_dblock *)(iobuf[d] + ((2 + q)*sector_size));
 
-			rv = verify_dblock(token, bk);
+			checksum = dblock_checksum(bk_end);
+
+			paxos_dblock_in(bk_end, &bk_in);
+			bk = &bk_in;
+
+			rv = verify_dblock(token, bk, checksum);
 			if (rv < 0)
 				continue;
 
@@ -557,12 +749,23 @@ static int run_ballot(struct task *task, struct token *token, int num_hosts,
 			continue;
 		free(iobuf[d]);
 	}
-	return error;
-}
 
-uint32_t leader_checksum(struct leader_record *lr)
-{
-	return crc32c((uint32_t)~1, (uint8_t *)lr, LEADER_CHECKSUM_LEN);
+	if (phase2 && (error < 0) &&
+	    ((error == SANLK_DBLOCK_READ) || (error == SANLK_DBLOCK_WRITE))) {
+		/*
+		 * After phase2 we might "win" the ballot even if we don't complete it
+		 * because another host could could pick and commit our dblock values.
+		 * If we abort the acquire, but are granted the lease, this would leave
+		 * us owning the lease on disk.  With this flag, the release path will
+		 * try to ensure we are not and do not become the lease owner.
+		 */
+		token->flags |= T_RETRACT_PAXOS;
+
+		log_errot(token, "ballot %llu retract error %d",
+			  (unsigned long long)next_lver, error);
+	}
+
+	return error;
 }
 
 static void log_leader_error(int result,
@@ -607,10 +810,11 @@ static void log_leader_error(int result,
 static int verify_leader(struct token *token,
 			 struct sync_disk *disk,
 			 struct leader_record *lr,
+			 uint32_t checksum,
 			 const char *caller)
 {
+	struct leader_record leader_end;
 	struct leader_record leader_rr;
-	uint32_t sum;
 	int result, rv;
 
 	if (lr->magic != PAXOS_DISK_MAGIC) {
@@ -656,11 +860,9 @@ static int verify_leader(struct token *token,
 		goto fail;
 	}
 
-	sum = leader_checksum(lr);
-
-	if (lr->checksum != sum) {
+	if (lr->checksum != checksum) {
 		log_errot(token, "verify_leader wrong checksum %x %x %s",
-			  lr->checksum, sum, disk->path);
+			  lr->checksum, checksum, disk->path);
 		result = SANLK_LEADER_CHECKSUM;
 		goto fail;
 	}
@@ -670,17 +872,28 @@ static int verify_leader(struct token *token,
  fail:
 	log_leader_error(result, token, disk, lr, caller);
 
-	memset(&leader_rr, 0, sizeof(leader_rr));
+	memset(&leader_end, 0, sizeof(struct leader_record));
 
-	rv = read_sectors(disk, 0, 1, (char *)&leader_rr,
+	rv = read_sectors(disk, 0, 1, (char *)&leader_end,
 			  sizeof(struct leader_record),
-			  NULL, "paxos_verify");
+			  NULL, 1, "paxos_verify");
+
+	leader_record_in(&leader_end, &leader_rr);
 
 	log_leader_error(rv, token, disk, &leader_rr, "paxos_verify");
 
 	return result;
 }
 
+int paxos_verify_leader(struct token *token,
+			 struct sync_disk *disk,
+			 struct leader_record *lr,
+			 uint32_t checksum,
+			 const char *caller)
+{
+	return verify_leader(token, disk, lr, checksum, caller);
+}
+
 static int leaders_match(struct leader_record *a, struct leader_record *b)
 {
 	if (!memcmp(a, b, LEADER_COMPARE_LEN))
@@ -688,21 +901,82 @@ static int leaders_match(struct leader_record *a, struct leader_record *b)
 	return 0;
 }
 
+/* read the lockspace name and resource name given the disk location */
+
+int paxos_read_resource(struct task *task,
+			struct token *token,
+			struct sanlk_resource *res)
+{
+	struct leader_record leader;
+	uint32_t checksum;
+	int rv;
+
+	memset(&leader, 0, sizeof(struct leader_record));
+
+	rv = read_leader(task, token, &token->disks[0], &leader, &checksum);
+	if (rv < 0)
+		return rv;
+
+	if (!res->lockspace_name[0])
+		memcpy(token->r.lockspace_name, leader.space_name, NAME_ID_SIZE);
+
+	if (!res->name[0])
+		memcpy(token->r.name, leader.resource_name, NAME_ID_SIZE);
+
+	rv = verify_leader(token, &token->disks[0], &leader, checksum, "read_resource");
+
+	if (rv == SANLK_OK) {
+		memcpy(res->lockspace_name, leader.space_name, NAME_ID_SIZE);
+		memcpy(res->name, leader.resource_name, NAME_ID_SIZE);
+		res->lver = leader.lver;
+	}
+
+	return rv;
+}
+
+int paxos_read_buf(struct task *task,
+		   struct token *token,
+		   char **buf_out)
+{
+	char *iobuf, **p_iobuf;
+	struct sync_disk *disk = &token->disks[0];
+	int rv, iobuf_len;
+
+	iobuf_len = direct_align(disk);
+	if (iobuf_len < 0)
+		return iobuf_len;
+
+	p_iobuf = &iobuf;
+
+	rv = posix_memalign((void *)p_iobuf, getpagesize(), iobuf_len);
+	if (rv)
+		return rv;
+
+	memset(iobuf, 0, iobuf_len);
+
+	rv = read_iobuf(disk->fd, disk->offset, iobuf, iobuf_len, task, token->io_timeout);
+
+	*buf_out = iobuf;
+
+	return rv;
+}
+
 static int _leader_read_one(struct task *task,
 			    struct token *token,
 			    struct leader_record *leader_ret,
 			    const char *caller)
 {
 	struct leader_record leader;
+	uint32_t checksum;
 	int rv;
 
 	memset(&leader, 0, sizeof(struct leader_record));
 
-	rv = read_leader(task, &token->disks[0], &leader);
+	rv = read_leader(task, token, &token->disks[0], &leader, &checksum);
 	if (rv < 0)
 		return rv;
 
-	rv = verify_leader(token, &token->disks[0], &leader, caller);
+	rv = verify_leader(token, &token->disks[0], &leader, checksum, caller);
 
 	/* copy what we read even if verify finds a problem */
 
@@ -719,6 +993,7 @@ static int _leader_read_num(struct task *task,
 {
 	struct leader_record leader;
 	struct leader_record *leaders;
+	uint32_t checksum;
 	int *leader_reps;
 	int leaders_len, leader_reps_len;
 	int num_reads;
@@ -751,11 +1026,11 @@ static int _leader_read_num(struct task *task,
 	num_reads = 0;
 
 	for (d = 0; d < num_disks; d++) {
-		rv = read_leader(task, &token->disks[d], &leaders[d]);
+		rv = read_leader(task, token, &token->disks[d], &leaders[d], &checksum);
 		if (rv < 0)
 			continue;
 
-		rv = verify_leader(token, &token->disks[d], &leaders[d], caller);
+		rv = verify_leader(token, &token->disks[d], &leaders[d], checksum, caller);
 		if (rv < 0)
 			continue;
 
@@ -843,10 +1118,14 @@ static int _lease_read_one(struct task *task,
 			   int *max_q,
 			   const char *caller)
 {
+	struct leader_record leader_end;
+	struct paxos_dblock our_dblock_end;
+	struct paxos_dblock bk;
 	char *iobuf, **p_iobuf;
 	uint32_t host_id = token->host_id;
 	uint32_t sector_size = disk->sector_size;
-	struct paxos_dblock *bk;
+	uint32_t checksum;
+	struct paxos_dblock *bk_end;
 	uint64_t tmp_mbal = 0;
 	int q, tmp_q = -1, rv, iobuf_len;
 
@@ -862,27 +1141,36 @@ static int _lease_read_one(struct task *task,
 
 	memset(iobuf, 0, iobuf_len);
 
-	rv = read_iobuf(disk->fd, disk->offset, iobuf, iobuf_len, task);
+	rv = read_iobuf(disk->fd, disk->offset, iobuf, iobuf_len, task, token->io_timeout);
 	if (rv < 0)
 		goto out;
 
-	memcpy(leader_ret, iobuf, sizeof(struct leader_record));
+	memcpy(&leader_end, iobuf, sizeof(struct leader_record));
 
-	memcpy(our_dblock, iobuf + ((host_id + 1) * sector_size), sizeof(struct paxos_dblock));
+	checksum = leader_checksum(&leader_end);
 
-	rv = verify_leader(token, disk, leader_ret, caller);
+	leader_record_in(&leader_end, leader_ret);
+
+	memcpy(&our_dblock_end, iobuf + ((host_id + 1) * sector_size), sizeof(struct paxos_dblock));
+	paxos_dblock_in(&our_dblock_end, our_dblock);
+
+	rv = verify_leader(token, disk, leader_ret, checksum, caller);
 	if (rv < 0)
 		goto out;
 
 	for (q = 0; q < leader_ret->num_hosts; q++) {
-		bk = (struct paxos_dblock *)(iobuf + ((2 + q) * sector_size));
+		bk_end = (struct paxos_dblock *)(iobuf + ((2 + q) * sector_size));
 
-		rv = verify_dblock(token, bk);
+		checksum = dblock_checksum(bk_end);
+
+		paxos_dblock_in(bk_end, &bk);
+
+		rv = verify_dblock(token, &bk, checksum);
 		if (rv < 0)
 			goto out;
 
-		if (!tmp_mbal || bk->mbal > tmp_mbal) {
-			tmp_mbal = bk->mbal;
+		if (!tmp_mbal || bk.mbal > tmp_mbal) {
+			tmp_mbal = bk.mbal;
 			tmp_q = q;
 		}
 	}
@@ -913,7 +1201,7 @@ static int _lease_read_num(struct task *task,
 	int *leader_reps;
 	int num_disks = token->r.num_disks;
 	int leaders_len, leader_reps_len;
-	int i, d, rv, found, num_reads, q_one, tmp_q = -1;
+	int i, d, rv = 0, found, num_reads, q_one, tmp_q = -1;
 
 	leaders_len = num_disks * sizeof(struct leader_record);
 	leader_reps_len = num_disks * sizeof(int);
@@ -1044,26 +1332,33 @@ static int write_new_leader(struct task *task,
 {
 	int num_disks = token->r.num_disks;
 	int num_writes = 0;
-	int error = SANLK_OK;
-	int rv = 0, d;
+	int timeout = 0;
+	int rv = 0;
+	int d;
 
 	for (d = 0; d < num_disks; d++) {
-		rv = write_leader(task, &token->disks[d], nl);
-		if (rv < 0)
+		rv = write_leader(task, token, &token->disks[d], nl);
+		if (rv == SANLK_AIO_TIMEOUT)
+			timeout = 1;
+		if (rv < 0) 
 			continue;
 		num_writes++;
 	}
 
 	if (!majority_disks(num_disks, num_writes)) {
-		log_errot(token, "%s write_new_leader error %d owner %llu %llu %llu",
-			  caller, rv,
+		log_errot(token, "%s write_new_leader error %d timeout %d owner %llu %llu %llu",
+			  caller, rv, timeout,
 			  (unsigned long long)nl->owner_id,
 			  (unsigned long long)nl->owner_generation,
 			  (unsigned long long)nl->timestamp);
-		error = SANLK_LEADER_WRITE;
+		if (timeout)
+			return SANLK_AIO_TIMEOUT;
+		if (rv < 0)
+			return rv;
+		return SANLK_LEADER_WRITE;
 	}
 
-	return error;
+	return SANLK_OK;
 }
 
 /*
@@ -1107,6 +1402,7 @@ int paxos_lease_acquire(struct task *task,
 	struct leader_record tmp_leader;
 	struct leader_record new_leader;
 	struct paxos_dblock dblock;
+	struct paxos_dblock owner_dblock;
 	struct host_status hs;
 	uint64_t wait_start, now;
 	uint64_t last_timestamp;
@@ -1117,9 +1413,13 @@ int paxos_lease_acquire(struct task *task,
 	int copy_cur_leader = 0;
 	int disk_open = 0;
 	int error, rv, us;
+	int other_io_timeout, other_host_dead_seconds;
+
+	memset(&dblock, 0, sizeof(dblock)); /* shut up compiler */
 
 	log_token(token, "paxos_acquire begin %x %llu %d",
 		  flags, (unsigned long long)acquire_lver, new_num_hosts);
+
  restart:
 
 	error = paxos_lease_read(task, token, &cur_leader, &max_mbal, "paxos_acquire");
@@ -1148,13 +1448,33 @@ int paxos_lease_acquire(struct task *task,
 
 	if (cur_leader.owner_id == token->host_id &&
 	    cur_leader.owner_generation == token->host_generation) {
-		log_token(token, "paxos_acquire already owner id %llu gen %llu",
+		log_token(token, "paxos_acquire owner %llu %llu %llu is already local %llu %llu",
+			  (unsigned long long)cur_leader.owner_id,
+			  (unsigned long long)cur_leader.owner_generation,
+			  (unsigned long long)cur_leader.timestamp,
 			  (unsigned long long)token->host_id,
 			  (unsigned long long)token->host_generation);
 		copy_cur_leader = 1;
 		goto run;
 	}
 
+	/*
+	 * We were the last host to hold this lease, but in a previous
+	 * lockspace generation in which we didn't cleanly release the
+	 * paxos lease.
+	 */
+
+	if (cur_leader.owner_id == token->host_id &&
+	    cur_leader.owner_generation < token->host_generation) {
+		log_token(token, "paxos_acquire owner %llu %llu %llu was old local new is %llu",
+			  (unsigned long long)cur_leader.owner_id,
+			  (unsigned long long)cur_leader.owner_generation,
+			  (unsigned long long)cur_leader.timestamp,
+			  (unsigned long long)token->host_generation);
+		copy_cur_leader = 1;
+		goto run;
+	}
+
 	/*
 	 * Check if current owner is alive based on its host_id renewals.
 	 * If the current owner has been dead long enough we can assume that
@@ -1204,7 +1524,7 @@ int paxos_lease_acquire(struct task *task,
 		  (unsigned long long)wait_start);
 
 	while (1) {
-		error = delta_lease_leader_read(task, &host_id_disk,
+		error = delta_lease_leader_read(task, token->io_timeout, &host_id_disk,
 						cur_leader.space_name,
 						cur_leader.owner_id,
 						&host_id_leader,
@@ -1258,9 +1578,19 @@ int paxos_lease_acquire(struct task *task,
 			goto skip_live_check;
 		}
 
-		/* the owner is renewing its host_id so it's alive */
+		/*
+		 * Check if the owner is alive:
+		 *
+		 * 1. We just read the delta lease of the owner (host_id_leader).
+		 * If that has a newer timestamp than the timestamp last seen by
+		 * our own renewal thread (last_timestamp), then the owner is alive.
+		 *
+		 * 2. If our own renewal thread saw the owner's timestamp change
+		 * the last time it was checked, then consider the owner to be alive.
+		 */
 
-		if (host_id_leader.timestamp != last_timestamp) {
+		if ((host_id_leader.timestamp != last_timestamp) ||
+		    (hs.last_live && (hs.last_check == hs.last_live))) {
 			if (flags & PAXOS_ACQUIRE_QUIET_FAIL) {
 				log_token(token, "paxos_acquire owner %llu "
 					  "delta %llu %llu %llu alive",
@@ -1277,18 +1607,46 @@ int paxos_lease_acquire(struct task *task,
 					  (unsigned long long)host_id_leader.timestamp);
 			}
 			memcpy(leader_ret, &cur_leader, sizeof(struct leader_record));
+
+			/* It's possible that the live owner has released the
+			   lease, but its release was clobbered by another host
+			   that was running the ballot with it and wrote it as
+			   the owner.  If the leader writer was not the owner,
+			   check if the owner's dblock is cleared.  If so, then
+			   the owner released the lease and we can run a
+			   ballot.  Comparing the write_id and owner_id is not
+			   required; we could always read the owner dblock
+			   here, but comparing the writer and owner can
+			   eliminate many unnecessary dblock reads. */
+
+			if (cur_leader.write_id != cur_leader.owner_id) {
+				rv = read_dblock(task, token, &token->disks[0],
+						 cur_leader.owner_id, &owner_dblock);
+				if (!rv && !owner_dblock.lver) {
+					/* not an error, but interesting to see */
+					log_errot(token, "paxos_acquire owner %llu %llu %llu writer %llu owner dblock free",
+						  (unsigned long long)cur_leader.owner_id,
+						  (unsigned long long)cur_leader.owner_generation,
+						  (unsigned long long)cur_leader.timestamp,
+						  (unsigned long long)cur_leader.write_id);
+					goto run;
+				}
+			}
+
 			error = SANLK_ACQUIRE_IDLIVE;
 			goto out;
 		}
 
-
-		/* if the owner hasn't renewed its host_id lease for
-		   host_dead_seconds then its watchdog should have fired
-		   by now */
+		/* If the owner hasn't renewed its host_id lease for
+		   host_dead_seconds then its watchdog should have fired by
+		   now. */
 
 		now = monotime();
 
-		if (now - wait_start > task->host_dead_seconds) {
+		other_io_timeout = hs.io_timeout;
+		other_host_dead_seconds = calc_host_dead_seconds(other_io_timeout);
+
+		if (now - wait_start > other_host_dead_seconds) {
 			log_token(token, "paxos_acquire owner %llu %llu %llu "
 				  "delta %llu %llu %llu dead %llu-%llu>%d",
 				  (unsigned long long)cur_leader.owner_id,
@@ -1299,10 +1657,19 @@ int paxos_lease_acquire(struct task *task,
 				  (unsigned long long)host_id_leader.timestamp,
 				  (unsigned long long)now,
 				  (unsigned long long)wait_start,
-				  task->host_dead_seconds);
+				  other_host_dead_seconds);
 			goto run;
 		}
 
+		if (flags & PAXOS_ACQUIRE_OWNER_NOWAIT) {
+			log_token(token, "paxos_acquire owner %llu %llu %llu no wait",
+				  (unsigned long long)cur_leader.owner_id,
+				  (unsigned long long)cur_leader.owner_generation,
+				  (unsigned long long)cur_leader.timestamp);
+			error = SANLK_ACQUIRE_OWNED_RETRY;
+			goto out;
+		}
+
  skip_live_check:
 		/* TODO: test with sleep(2) here */
 		sleep(1);
@@ -1475,15 +1842,24 @@ int paxos_lease_acquire(struct task *task,
 			new_leader.flags &= ~LFL_SHORT_HOLD;
 	}
 
-	new_leader.checksum = leader_checksum(&new_leader);
+	new_leader.checksum = 0; /* set after leader_record_out */
 
 	error = write_new_leader(task, token, &new_leader, "paxos_acquire");
-	if (error < 0)
+	if (error < 0) {
+		/* See comment in run_ballot about this flag. */
+		token->flags |= T_RETRACT_PAXOS;
+		memcpy(leader_ret, &new_leader, sizeof(struct leader_record));
 		goto out;
+	}
 
 	if (new_leader.owner_id != token->host_id) {
 		/* not a problem, but interesting to see, so use log_error */
 
+		/* It's possible that we commit an outdated owner id/gen here.
+		   If we go back to the top and retry, we may find that the
+		   owner host_id is alive but with a newer generation, and
+		   we'd be able to get the lease by running the ballot again. */
+
 		log_errot(token, "ballot %llu commit other owner %llu %llu %llu",
 			  (unsigned long long)new_leader.lver,
 			  (unsigned long long)new_leader.owner_id,
@@ -1524,7 +1900,7 @@ int paxos_lease_renew(struct task *task,
 	for (d = 0; d < token->r.num_disks; d++) {
 		memset(&new_leader, 0, sizeof(struct leader_record));
 
-		rv = read_leader(task, &token->disks[d], &new_leader);
+		rv = read_leader(task, token, &token->disks[d], &new_leader);
 		if (rv < 0)
 			continue;
 
@@ -1536,7 +1912,7 @@ int paxos_lease_renew(struct task *task,
 	}
 
 	new_leader.timestamp = monotime();
-	new_leader.checksum = leader_checksum(&new_leader);
+	new_leader.checksum = 0; /* set after leader_record_out */
 
 	error = write_new_leader(task, token, &new_leader);
 	if (error < 0)
@@ -1550,10 +1926,12 @@ int paxos_lease_renew(struct task *task,
 
 int paxos_lease_release(struct task *task,
 			struct token *token,
+			struct sanlk_resource *resrename,
 		        struct leader_record *leader_last,
 		        struct leader_record *leader_ret)
 {
 	struct leader_record leader;
+	struct leader_record *last;
 	int error;
 
 	error = paxos_lease_leader_read(task, token, &leader, "paxos_release");
@@ -1562,24 +1940,42 @@ int paxos_lease_release(struct task *task,
 		goto out;
 	}
 
-	if (leader.lver != leader_last->lver) {
+	/*
+	 * Used when the caller does not know who the owner is, but
+	 * wants to ensure it is not the owner.
+	 */
+	if (!leader_last)
+		last = &leader;
+	else
+		last = leader_last;
+
+	if (leader.lver != last->lver) {
 		log_errot(token, "paxos_release %llu other lver %llu",
-			  (unsigned long long)leader_last->lver,
+			  (unsigned long long)last->lver,
 			  (unsigned long long)leader.lver);
 		return SANLK_RELEASE_LVER;
 	}
 
-	if (leader.owner_id != token->host_id ||
-	    leader.owner_generation != token->host_generation) {
-		log_errot(token, "paxos_release %llu other owner %llu %llu %llu",
-			  (unsigned long long)leader_last->lver,
+	if (leader.timestamp == LEASE_FREE) {
+		log_errot(token, "paxos_release %llu already free %llu %llu %llu",
+			  (unsigned long long)last->lver,
 			  (unsigned long long)leader.owner_id,
 			  (unsigned long long)leader.owner_generation,
 			  (unsigned long long)leader.timestamp);
 		return SANLK_RELEASE_OWNER;
 	}
 
-	if (memcmp(&leader, leader_last, sizeof(struct leader_record))) {
+	if (leader.owner_id != token->host_id ||
+	    leader.owner_generation != token->host_generation) {
+		log_errot(token, "paxos_release %llu other owner %llu %llu %llu",
+			  (unsigned long long)last->lver,
+			  (unsigned long long)leader.owner_id,
+			  (unsigned long long)leader.owner_generation,
+			  (unsigned long long)leader.timestamp);
+		return SANLK_RELEASE_OWNER;
+	}
+
+	if (memcmp(&leader, last, sizeof(struct leader_record))) {
 		/*
 		 * This will happen when two hosts finish the same ballot
 		 * successfully, the second commiting the same inp values
@@ -1593,25 +1989,28 @@ int paxos_lease_release(struct task *task,
 		 */
 		log_errot(token, "paxos_release %llu leader different "
 			  "write %llu %llu %llu vs %llu %llu %llu",
-			  (unsigned long long)leader_last->lver,
-			  (unsigned long long)leader_last->write_id,
-			  (unsigned long long)leader_last->write_generation,
-			  (unsigned long long)leader_last->write_timestamp,
+			  (unsigned long long)last->lver,
+			  (unsigned long long)last->write_id,
+			  (unsigned long long)last->write_generation,
+			  (unsigned long long)last->write_timestamp,
 			  (unsigned long long)leader.write_id,
 			  (unsigned long long)leader.write_generation,
 			  (unsigned long long)leader.write_timestamp);
 		/*
-		log_leader_error(0, token, &token->disks[0], leader_last, "paxos_release");
+		log_leader_error(0, token, &token->disks[0], last, "paxos_release");
 		log_leader_error(0, token, &token->disks[0], &leader, "paxos_release");
 		*/
 	}
 
+	if (resrename)
+		memcpy(leader.resource_name, resrename->name, NAME_ID_SIZE);
+
 	leader.timestamp = LEASE_FREE;
 	leader.write_id = token->host_id;
 	leader.write_generation = token->host_generation;
 	leader.write_timestamp = monotime();
 	leader.flags &= ~LFL_SHORT_HOLD;
-	leader.checksum = leader_checksum(&leader);
+	leader.checksum = 0; /* set after leader_record_out */
 
 	error = write_new_leader(task, token, &leader, "paxos_release");
 	if (error < 0)
@@ -1627,8 +2026,11 @@ int paxos_lease_init(struct task *task,
 		     int num_hosts, int max_hosts)
 {
 	char *iobuf, **p_iobuf;
-	struct leader_record *leader;
-	struct request_record *rr;
+	struct leader_record leader;
+	struct leader_record leader_end;
+	struct request_record rr;
+	struct request_record rr_end;
+	uint32_t checksum;
 	int iobuf_len;
 	int sector_size;
 	int align_size;
@@ -1640,6 +2042,15 @@ int paxos_lease_init(struct task *task,
 	if (!max_hosts)
 		max_hosts = DEFAULT_MAX_HOSTS;
 
+	if (max_hosts > DEFAULT_MAX_HOSTS)
+		return -E2BIG;
+
+	if (num_hosts > DEFAULT_MAX_HOSTS)
+		return -EINVAL;
+
+	if (num_hosts > max_hosts)
+		return -EINVAL;
+
 	sector_size = token->disks[0].sector_size;
 
 	align_size = direct_align(&token->disks[0]);
@@ -1659,24 +2070,38 @@ int paxos_lease_init(struct task *task,
 
 	memset(iobuf, 0, iobuf_len);
 
-	leader = (struct leader_record *)iobuf;
-	leader->magic = PAXOS_DISK_MAGIC;
-	leader->version = PAXOS_DISK_VERSION_MAJOR | PAXOS_DISK_VERSION_MINOR;
-	leader->sector_size = sector_size;
-	leader->num_hosts = num_hosts;
-	leader->max_hosts = max_hosts;
-	leader->timestamp = LEASE_FREE;
-	strncpy(leader->space_name, token->r.lockspace_name, NAME_ID_SIZE);
-	strncpy(leader->resource_name, token->r.name, NAME_ID_SIZE);
-	leader->checksum = leader_checksum(leader);
+	memset(&leader, 0, sizeof(leader));
+	leader.magic = PAXOS_DISK_MAGIC;
+	leader.version = PAXOS_DISK_VERSION_MAJOR | PAXOS_DISK_VERSION_MINOR;
+	leader.sector_size = sector_size;
+	leader.num_hosts = num_hosts;
+	leader.max_hosts = max_hosts;
+	leader.timestamp = LEASE_FREE;
+	strncpy(leader.space_name, token->r.lockspace_name, NAME_ID_SIZE);
+	strncpy(leader.resource_name, token->r.name, NAME_ID_SIZE);
+	leader.checksum = 0; /* set after leader_record_out */
 
-	rr = (struct request_record *)(iobuf + sector_size);
-	rr->magic = REQ_DISK_MAGIC;
-	rr->version = REQ_DISK_VERSION_MAJOR | REQ_DISK_VERSION_MINOR;
+	memset(&rr, 0, sizeof(rr));
+	rr.magic = REQ_DISK_MAGIC;
+	rr.version = REQ_DISK_VERSION_MAJOR | REQ_DISK_VERSION_MINOR;
+
+	leader_record_out(&leader, &leader_end);
+
+	/*
+	 * N.B. must compute checksum after the data has been byte swapped.
+	 */
+	checksum = leader_checksum(&leader_end);
+	leader.checksum = checksum;
+	leader_end.checksum = cpu_to_le32(checksum);
+
+	request_record_out(&rr, &rr_end);
+
+	memcpy(iobuf, &leader_end, sizeof(struct leader_record));
+	memcpy(iobuf + sector_size, &rr_end, sizeof(struct request_record));
 
 	for (d = 0; d < token->r.num_disks; d++) {
 		rv = write_iobuf(token->disks[d].fd, token->disks[d].offset,
-				 iobuf, iobuf_len, task);
+				 iobuf, iobuf_len, task, token->io_timeout);
 
 		if (rv == SANLK_AIO_TIMEOUT)
 			aio_timeout = 1;
