@@ -12,6 +12,7 @@
 #define PAXOS_ACQUIRE_FORCE		0x00000001
 #define PAXOS_ACQUIRE_QUIET_FAIL	0x00000002
 #define PAXOS_ACQUIRE_SHARED		0x00000004
+#define PAXOS_ACQUIRE_OWNER_NOWAIT	0x00000008
 
 uint32_t leader_checksum(struct leader_record *lr);
 
@@ -29,6 +30,7 @@ int paxos_lease_acquire(struct task *task,
 
 int paxos_lease_release(struct task *task,
 			struct token *token,
+			struct sanlk_resource *resrename,
 			struct leader_record *leader_last,
 			struct leader_record *leader_ret);
 
@@ -42,4 +44,26 @@ int paxos_lease_request_read(struct task *task, struct token *token,
 int paxos_lease_request_write(struct task *task, struct token *token,
                               struct request_record *rr);
 
+int paxos_read_resource(struct task *task,
+			struct token *token,
+			struct sanlk_resource *res);
+
+int paxos_read_buf(struct task *task,
+                   struct token *token,
+                   char **buf_out);
+
+int paxos_verify_leader(struct token *token,
+                         struct sync_disk *disk,
+                         struct leader_record *lr,
+			 uint32_t checksum,
+                         const char *caller);
+
+int paxos_erase_dblock(struct task *task,
+                       struct token *token,
+                       uint64_t host_id);
+
+int paxos_lease_leader_clobber(struct task *task,
+                               struct token *token,
+                               struct leader_record *leader,
+                               const char *caller);
 #endif
